@@ -1,4 +1,4 @@
-#line 2 "perl-libpng.c.tmpl"
+#line 2 "tmpl/perl-libpng.c.tmpl"
 
 /* Coding style note: use "color" in code & variables, not "colour",
    as per libpng. 2020-12-09 08:03:03 */
@@ -167,6 +167,21 @@ perl_png_warning_fn (png_structp png_ptr, png_const_charp warning_msg)
                                                  
 
 
+/* The following is a debugging construction which is usually "off". */
+
+#if 0
+#define GABBLE_IN(x)			\
+    fprintf (stderr, "Get #%d of %s\n", \
+	     png->memory_gets, #x)
+
+#define GABBLE_OUT(x)				\
+    fprintf (stderr, "Free #%d of %s\n",	\
+	     png->memory_gets, #x)
+#else
+#define GABBLE_IN(x)
+#define GABBLE_OUT(x)
+#endif /* 0 */
+
 /* Get memory using the following in order to keep count of the number
    of objects in use at the end of execution, to ensure that there are
    no memory leaks. All allocation is done via Newxz ("calloc") rather
@@ -175,6 +190,7 @@ perl_png_warning_fn (png_structp png_ptr, png_const_charp warning_msg)
 #define GET_MEMORY(thing, number, type) {	\
         Newxz (thing, number, type);		\
         png->memory_gets++;                     \
+	GABBLE_IN(thing);			\
     }
 
 /* Free memory using the following in order to keep count of the
@@ -183,6 +199,7 @@ perl_png_warning_fn (png_structp png_ptr, png_const_charp warning_msg)
 #define PERL_PNG_FREE(thing) {   \
         png->memory_gets--;      \
         Safefree (thing);	 \
+	GABBLE_OUT(thing);	 \
     }
 
 static perl_libpng_t *
@@ -510,37 +527,48 @@ perl_png_get_text (perl_libpng_t * png)
 #endif
 }
 
-/* Set a PNG text from "chunk". The return value is true if
-   successful. */
+/* The macro which SvPV consists of fouls things up with the STRLEN
+   pointer if we try to use a function. */
 
-static int
+#define SOFT_HASH_FETCH_PV(chunk, key) {			\
+	SV ** sv_ptr;						\
+	sv_ptr = hv_fetch (chunk, #key, strlen (#key), 0);	\
+	if (sv_ptr) {						\
+	    key = SvPV(*sv_ptr, key ## _length);		\
+	} else {						\
+	    key = 0;						\
+	    key ## _length = 0;					\
+	}							\
+    }
+
+/* Set a PNG text "text_out" from "chunk". */
+
+static void
 perl_png_set_text_from_hash (perl_libpng_t * png,
-                             png_text * text_out, HV * chunk)
+                             png_text * png_texts, int i, HV * chunk)
 {
+    png_text * text_out;
     int compression;
     char * key;
     STRLEN key_length;
-#ifdef PNG_iTXt_SUPPORTED
-    char * lang;
-    STRLEN lang_length;
-    char * lang_key;
-    STRLEN lang_key_length;
-#endif /* PNG_iTXt_SUPPORTED */
-    int is_itxt = 0;
-    char * text;
+    char * text = 0;
     STRLEN text_length;
-    /* The return value of this function. */
-    int ok = 1;
-    SV ** compression_sv_ptr;
+    SV ** c_sv_ptr;
+#ifdef PNG_iTXt_SUPPORTED
+    char * lang = 0;
+    STRLEN lang_length;
+    char * lang_key = 0;
+    STRLEN lang_key_length;
+    int is_itxt = 0;
+#endif /* PNG_iTXt_SUPPORTED */
 
-    MESSAGE ("Putting it into something.");
+    text_out = & png_texts[i];
 
     /* Check the compression field of the chunk */
 
-    compression_sv_ptr =
-	hv_fetch (chunk, "compression", strlen ("compression"), 0);
-    if (compression_sv_ptr) {
-	compression = SvIV (* compression_sv_ptr);
+    c_sv_ptr = hv_fetch (chunk, "compression", strlen ("compression"), 0);
+    if (c_sv_ptr) {
+	compression = SvIV (* c_sv_ptr);
     }
     else {
 	MESSAGE ("Using default compression PNG_TEXT_COMPRESSION_NONE");
@@ -548,55 +576,51 @@ perl_png_set_text_from_hash (perl_libpng_t * png,
     }
     switch (compression) {
     case PNG_TEXT_COMPRESSION_NONE:
-        break;
     case PNG_TEXT_COMPRESSION_zTXt:
         break;
+#ifdef PNG_iTXt_SUPPORTED
     case PNG_ITXT_COMPRESSION_NONE:
-        is_itxt = 1;
-        break;
     case PNG_ITXT_COMPRESSION_zTXt: 
         is_itxt = 1;
         break;
+#endif /* PNG_iTXt_SUPPORTED */
     default:
-        ok = 0;
-        fprintf (stderr, "Unknown compression %d\n", 
-                 compression);
-        return 0;
-        break;
+	PERL_PNG_FREE(png_texts);
+	croak ("Unknown compression %d", compression);
     }
+    text_out->compression = compression;
 
     MESSAGE ("Getting key.");
-    HASH_FETCH_PV (chunk, key);
-    if (key_length < 1 || key_length > 79) {
-        /* Key is too long or empty */
-        MESSAGE ("Bad length of key.");
-
-        ok = 0;
-        return 0;
+    SOFT_HASH_FETCH_PV (chunk, key);
+    if (key == 0) {
+	PERL_PNG_FREE(png_texts);
+	croak ("Text chunk %d has no 'key' field", i);
     }
-    MESSAGE ("Getting text.");
-    HASH_FETCH_PV (chunk, text);
-    if (ok) {
-        MESSAGE ("Copying.");
-        text_out->compression = compression;
-        text_out->key = (char *) key;
-        text_out->text = (char *) text;
-        text_out->text_length = text_length;
+    if (key_length < 1) {
+	PERL_PNG_FREE(png_texts);
+	croak ("Text chunk %d key field is empty", i);
+    }
+    if (key_length > 79) {
+	PERL_PNG_FREE(png_texts);
+	croak ("Text chunk %d key field is too long %d > 79",
+	       i, (int) key_length);
+    }
+    text_out->key = (char *) key;
+    /* Libpng documentation says it is OK to send NULLs here. */
+    SOFT_HASH_FETCH_PV (chunk, text);
+    text_out->text = (char *) text;
+    text_out->text_length = text_length;
+    /* Set this in case it starts to be required by future versions of
+       libpng. */
+    text_out->itxt_length = text_length;
 #ifdef PNG_iTXt_SUPPORTED
-        if (is_itxt) {
-            HASH_FETCH_PV (chunk, lang);
-            HASH_FETCH_PV (chunk, lang_key);
-            text_out->lang = (char *) lang;
-            text_out->lang_key = (char *) lang_key;
-        }
-#endif
+    if (is_itxt) {
+	SOFT_HASH_FETCH_PV (chunk, lang);
+	text_out->lang = (char *) lang;
+	SOFT_HASH_FETCH_PV (chunk, lang_key);
+	text_out->lang_key = (char *) lang_key;
     }
-    else {
-            /* Compression method unknown. */
-        ;
-    }
-
-    return ok;
+#endif /* PNG_iTXt_SUPPORTED */
 }
 
 /* Set the text chunks in the PNG. This actually pushes text chunks
@@ -607,52 +631,43 @@ static void
 perl_png_set_text (perl_libpng_t * png, AV * text_chunks)
 {
     int num_text;
-    int num_ok = 0;
     int i;
     png_text * png_texts;
 
+#ifndef PNG_tEXt_SUPPORTED
+    UNSUPPORTED(tEXt);
+    return;
+#endif
+
     num_text = av_len (text_chunks) + 1;
     MESSAGE ("You have %d text chunks.\n", num_text);
     if (num_text <= 0) {
-        /* Complain to the user */
         return;
     }
+    /* This memory needs to be freed before we call "croak", hence we
+       cannot "croak" inside perl_png_set_text_from_hash but have to
+       use error return values, then free this, and then croak. */
     GET_MEMORY (png_texts, num_text, png_text);
     for (i = 0; i < num_text; i++) {
-        int ok = 0;
         SV ** chunk_pointer;
 
         MESSAGE ("Fetching chunk %d.\n", i);
         chunk_pointer = av_fetch (text_chunks, i, 0);
         if (! chunk_pointer) {
-            /* Complain */
-            MESSAGE ("Chunk pointer null.");
-            continue;
+	    PERL_PNG_FREE(png_texts);
+	    croak ("Null chunk pointer");
         }
         if (SvROK (* chunk_pointer) && 
             SvTYPE (SvRV (* chunk_pointer)) == SVt_PVHV) {
-            MESSAGE ("Looks like a hash.");
-            ok = perl_png_set_text_from_hash (png, & png_texts[num_ok],
-                                              (HV *) SvRV (* chunk_pointer));
-            if (ok) {
-                MESSAGE ("This chunk is OK.");
-                num_ok++;
-            }
-            else {
-                MESSAGE ("The chunk is not OK.");
-            }
+	    perl_png_set_text_from_hash (png, png_texts, i,
+					 (HV *) SvRV (* chunk_pointer));
         }
+	else {
+	    PERL_PNG_FREE(png_texts);
+	    croak ("Element %d of text_chunks is not a hash reference", i);
+	}
     }
-    if (num_ok > 0) {
-        MESSAGE ("Writing %d text chunks to your PNG.\n",
-                num_ok);
-        png_set_text (pngi, png_texts, num_ok);
-    }
-    else {
-        /* The user tried to set some text chunks in the image but
-           they were not allowed. */
-        warn ( "None of your text chunks was allowed");
-    }
+    png_set_text (pngi, png_texts, num_text);
     PERL_PNG_FREE (png_texts);
 }
 
@@ -726,44 +741,43 @@ perl_png_get_tIME (perl_libpng_t * png)
 static void
 perl_png_set_tIME (perl_libpng_t * png, SV * input_time)
 {
-    /* The PNG month and day fields shouldn't be equal to zero. 
-       See PNG specification "4.2.4.6. tIME Image last-modification time"
-    */
+    /* The PNG month and day fields shouldn't be equal to zero.
+       See PNG specification "4.2.4.6. tIME Image
+       last-modification time". */
     png_time mod_time = {0,1,1,0,0,0};
-    int time_ok = 0;
     if (input_time) {
 	SV * ref;
+	HV * time_hash;
         ref = SvRV(input_time);
-        if (ref && SvTYPE (ref) == SVt_PVHV) {
-            HV * time_hash = (HV *) SvRV (input_time);
-            MESSAGE ("Setting time from a hash.");
-#define SET_TIME(field) {                                               \
-                SV ** field_sv_ptr = hv_fetch (time_hash, #field,       \
-                                               strlen (#field), 0);     \
-                if (field_sv_ptr) {                                     \
-                    SV * field_sv = * field_sv_ptr;                     \
-                    MESSAGE ("OK for %s\n", #field);                    \
-                    mod_time.field = SvIV (field_sv);                   \
-                }                                                       \
-            }
-            SET_TIME(year);
-            SET_TIME(month);
-            SET_TIME(day);
-            SET_TIME(hour);
-            SET_TIME(minute);
-            SET_TIME(second);
-#undef SET_TIME    
-            time_ok = 1;
-        }
-    }
-    if (! time_ok) {
-        /* There is no valid time argument, so just set it to the time
-           now, according to the system clock. */
-        time_t now;
+        if (! ref || SvTYPE (ref) != SVt_PVHV) {
+	    croak ("Argument to set_tIME should be a hash reference");
+	}
+	time_hash = (HV *) ref;
 
-        MESSAGE ("The modification time doesn't look OK so I am going to set the modification time to the time now instead.");
-        now = time (0);
-        png_convert_from_time_t (& mod_time, now);
+	MESSAGE ("Setting time from a hash.");
+#define SET_TIME(field) {					\
+	    SV ** field_sv_ptr = hv_fetch (time_hash, #field,	\
+					   strlen (#field), 0);	\
+	    if (field_sv_ptr) {					\
+		SV * field_sv = * field_sv_ptr;			\
+		MESSAGE ("OK for %s\n", #field);		\
+		mod_time.field = SvIV (field_sv);		\
+	    }							\
+	}
+	SET_TIME(year);
+	SET_TIME(month);
+	SET_TIME(day);
+	SET_TIME(hour);
+	SET_TIME(minute);
+	SET_TIME(second);
+#undef SET_TIME    
+    }
+    else {
+	/* Use the current time. */
+	time_t now;
+	
+	now = time (0);
+	png_convert_from_time_t (& mod_time, now);
     }
     png_set_tIME (pngi, & mod_time);
 }
@@ -1053,6 +1067,8 @@ const char * perl_png_color_type_name (int color_type)
     return name;
 }
 
+#undef PERL_PNG_COLOR_TYPE
+
 #define PERL_PNG_TEXT_COMP(x,y)                  \
     case PNG_ ## x ## _COMPRESSION_ ## y:        \
     name = #x "_" #y;                            \
@@ -1064,18 +1080,21 @@ const char * perl_png_text_compression_name (int text_compression)
 {
     const char * name;
     switch (text_compression) {
+#ifdef PNG_tEXt_SUPPORTED
         PERL_PNG_TEXT_COMP(TEXT,NONE);
         PERL_PNG_TEXT_COMP(TEXT,zTXt);
+#ifdef PNG_iTXt_SUPPORTED
         PERL_PNG_TEXT_COMP(ITXT,NONE);
         PERL_PNG_TEXT_COMP(ITXT,zTXt);
+#endif /* iTXt */
+#endif /* PNG_tEXt_SUPPORTED */
     default:
-        /* Moan about not knowing this text compression type. */
+	warn ("Unknown compression type %d", text_compression);
         name = "";
     }
     return name;
 }
 
-#undef PERL_PNG_COLOR_TYPE
 
 /*  ____       _      _   _       
    |  _ \ __ _| | ___| |_| |_ ___ 
@@ -1247,7 +1266,7 @@ static void perl_png_set_bKGD (perl_libpng_t * png, HV * bKGD)
 }
 
 /* Get the pCAL (calibration of pixel values) chunk from a PNG
-   image. This currently does nothing. */
+   image. */
 
 static SV * perl_png_get_pCAL (perl_libpng_t * png)
 {
@@ -1447,7 +1466,7 @@ static SV * perl_png_get_tRNS (perl_libpng_t * png)
     else {
 	HV * trans_hv;
 	trans_hv = newHV ();
-#line 1450 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 1469 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
         
 	HASH_STORE_IV_MEMBER (trans_hv, red, trans_values);
 	
@@ -1457,7 +1476,7 @@ static SV * perl_png_get_tRNS (perl_libpng_t * png)
 	
 	HASH_STORE_IV_MEMBER (trans_hv, gray, trans_values);
 	
-#line 1455 "perl-libpng.c.tmpl"
+#line 1474 "tmpl/perl-libpng.c.tmpl"
 	return newRV_noinc ((SV *) trans_hv);
     }
 #else
@@ -1512,7 +1531,7 @@ static void perl_png_set_tRNS (perl_libpng_t * png, SV * tRNS)
 	    /* unhandled error */
 	}
 	trans_hv = (HV *) SvRV (tRNS);
-#line 1515 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 1534 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
         
 	    HASH_FETCH_IV_MEMBER (trans_hv, red, (& trans_values));
 	
@@ -1522,7 +1541,7 @@ static void perl_png_set_tRNS (perl_libpng_t * png, SV * tRNS)
 	
 	    HASH_FETCH_IV_MEMBER (trans_hv, gray, (& trans_values));
 	
-#line 1514 "perl-libpng.c.tmpl"
+#line 1533 "tmpl/perl-libpng.c.tmpl"
 	num_trans = 1;
 	png_set_tRNS (pngi, trans, num_trans, & trans_values); 
     }
@@ -1550,14 +1569,14 @@ perl_png_spalette_to_hv (png_sPLT_tp spalette)
 	png_sPLT_entry * entry;
 	entry = spalette->entries + i;
 	perl_entry = newHV ();
-#line 1553 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 1572 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
 		HASH_STORE_IV_MEMBER (perl_entry, red, entry);
 		HASH_STORE_IV_MEMBER (perl_entry, green, entry);
 		HASH_STORE_IV_MEMBER (perl_entry, blue, entry);
 		HASH_STORE_IV_MEMBER (perl_entry, alpha, entry);
 		HASH_STORE_IV_MEMBER (perl_entry, frequency, entry);
 	
-#line 1546 "perl-libpng.c.tmpl"
+#line 1565 "tmpl/perl-libpng.c.tmpl"
 	av_push (entries, newRV_noinc ((SV *) perl_entry));
     }
     HASH_STORE_AV (perl_spalette, entries);
@@ -1661,14 +1680,14 @@ static void perl_png_set_sPLT (perl_libpng_t * png, AV * sPLT_entries)
 	    }
 	    MESSAGE ("Copying entry %d", j);
 	    e = entry->entries + j;
-#line 1664 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 1683 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
  	    	    HASH_FETCH_IV_MEMBER (perl_entry, red, e);
 	    	    HASH_FETCH_IV_MEMBER (perl_entry, green, e);
 	    	    HASH_FETCH_IV_MEMBER (perl_entry, blue, e);
 	    	    HASH_FETCH_IV_MEMBER (perl_entry, alpha, e);
 	    	    HASH_FETCH_IV_MEMBER (perl_entry, frequency, e);
 	    
-#line 1654 "perl-libpng.c.tmpl"
+#line 1673 "tmpl/perl-libpng.c.tmpl"
 	}
     }
     MESSAGE ("Setting the entries");
@@ -1781,7 +1800,7 @@ static SV * perl_png_get_sBIT (perl_libpng_t * png)
 	if (status != PNG_INFO_sBIT) {
 	    /* error */
 	}
-#line 1784 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 1803 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
         
 	HASH_STORE_IV_MEMBER (sig_bit, red, colors);
 	
@@ -1793,7 +1812,7 @@ static SV * perl_png_get_sBIT (perl_libpng_t * png)
 	
 	HASH_STORE_IV_MEMBER (sig_bit, alpha, colors);
 	
-#line 1771 "perl-libpng.c.tmpl"
+#line 1790 "tmpl/perl-libpng.c.tmpl"
         return newRV_noinc ((SV *) sig_bit);
     }
     return & PL_sv_undef;
@@ -1975,9 +1994,9 @@ static SV * perl_png_get_valid (perl_libpng_t * png)
     valid = png_get_valid (pngi, 0xFFFFFFFF);
 #define V(x) \
     (void) hv_store (perl_valid, #x, strlen (#x), newSViv (valid & PNG_INFO_ ## x), 0)
-#line 1978 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 1997 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
 V(bKGD);V(cHRM);V(gAMA);V(hIST);V(iCCP);V(IDAT);V(oFFs);V(pCAL);V(pHYs);V(PLTE);V(sBIT);V(sCAL);V(sPLT);V(sRGB);V(tIME);V(tRNS);
-#line 1955 "perl-libpng.c.tmpl"
+#line 1974 "tmpl/perl-libpng.c.tmpl"
 #undef V
 
     return newRV_noinc ((SV *) perl_valid);
@@ -2351,7 +2370,7 @@ static void perl_png_set_unknown_chunks (perl_libpng_t * png, AV * chunk_list)
 
 int perl_png_libpng_supports (const char * what)
 {
-#line 2354 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 2373 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
     if (strcmp (what, "iTXt") == 0) {
 #ifdef PNG_iTXt_SUPPORTED
         return 1;
@@ -2520,7 +2539,7 @@ int perl_png_libpng_supports (const char * what)
         return 0;
 #endif /* WRITE_CUSTOMIZE_COMPRESSION */
     }
-#line 2343 "perl-libpng.c.tmpl"
+#line 2362 "tmpl/perl-libpng.c.tmpl"
 
     /* sCAL is a special case. */
 
@@ -2626,7 +2645,7 @@ static SV * perl_png_get_cHRM (perl_libpng_t * png)
 #ifdef PNG_cHRM_SUPPORTED
     if (VALID (cHRM)) {
         HV * ice;
-#line 2629 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 2648 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
         double white_x;
         double white_y;
         double red_x;
@@ -2635,10 +2654,10 @@ static SV * perl_png_get_cHRM (perl_libpng_t * png)
         double green_y;
         double blue_x;
         double blue_y;
-#line 2453 "perl-libpng.c.tmpl"
+#line 2472 "tmpl/perl-libpng.c.tmpl"
         png_get_cHRM (pngi , & white_x, & white_y, & red_x, & red_y, & green_x, & green_y, & blue_x, & blue_y);
         ice = newHV ();
-#line 2641 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 2660 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
         (void) hv_store (ice, "white_x", strlen ("white_x"),
                          newSVnv (white_x), 0);
         (void) hv_store (ice, "white_y", strlen ("white_y"),
@@ -2655,7 +2674,7 @@ static SV * perl_png_get_cHRM (perl_libpng_t * png)
                          newSVnv (blue_x), 0);
         (void) hv_store (ice, "blue_y", strlen ("blue_y"),
                          newSVnv (blue_y), 0);
-#line 2462 "perl-libpng.c.tmpl"
+#line 2481 "tmpl/perl-libpng.c.tmpl"
         return newRV_noinc ((SV *) ice);
     }
     return & PL_sv_undef;
@@ -2670,7 +2689,7 @@ static void perl_png_set_cHRM (perl_libpng_t * png, HV * cHRM)
 {
 #ifdef PNG_cHRM_SUPPORTED
     SV ** key_sv_ptr;
-#line 2673 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 2692 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
     double white_x = 0.0;
     double white_y = 0.0;
     double red_x = 0.0;
@@ -2711,9 +2730,9 @@ static void perl_png_set_cHRM (perl_libpng_t * png, HV * cHRM)
     if (key_sv_ptr) {
         blue_y = SvNV (* key_sv_ptr);
     }
-#line 2487 "perl-libpng.c.tmpl"
+#line 2506 "tmpl/perl-libpng.c.tmpl"
     png_set_cHRM (pngi, white_x, white_y, red_x, red_y, green_x, green_y, blue_x, blue_y);
-#line 2490 "perl-libpng.c.tmpl"
+#line 2509 "tmpl/perl-libpng.c.tmpl"
 #else
     /* libpng was compiled without this option. */
     UNSUPPORTED(cHRM);
@@ -2747,7 +2766,7 @@ static void perl_png_copy_row_pointers (perl_libpng_t * png, SV * row_pointers)
     png_set_rows (pngi, png->row_pointers);
 }
 
-#line 2750 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
+#line 2769 "/usr/home/ben/projects/image-png-libpng/build/../perl-libpng.c"
 
 static int
 perl_png_get_image_width (perl_libpng_t * png)
@@ -2761,7 +2780,7 @@ perl_png_get_image_height (perl_libpng_t * png)
     return png_get_image_height (pngi);
 }
 
-#line 2532 "perl-libpng.c.tmpl"
+#line 2551 "tmpl/perl-libpng.c.tmpl"
 
 
 static void
